@@ -6,21 +6,25 @@
 
 #include <Rinternals.h>
 
-/* avoid protection issues with setAttrib where new symbols may trigger GC probelms */
+/* avoid protection issues with setAttrib where new symbols may trigger GC problems */
 static void setAttr(SEXP x, const char *name, SEXP val) {
     PROTECT(val);
     setAttrib(x, Rf_install(name), val);
     UNPROTECT(1);
 }
 
-/* add information attributes accorsing to the TIGG tags.
-   Only a somewhat randome set (albeit mostly baseline) is supported */
+/* add information attributes according to the TIFF tags.
+   Only a somewhat random set (albeit mostly baseline) is supported */
 static void TIFF_add_info(TIFF *tiff, SEXP res) {
     uint32 i32;
     uint16 i16;
     float f;
     char *c = 0;
 
+    if (TIFFGetField(tiff, TIFFTAG_IMAGEWIDTH, &i32))
+	setAttr(res, "width", ScalarInteger(i32));
+    if (TIFFGetField(tiff, TIFFTAG_IMAGELENGTH, &i32))
+	setAttr(res, "length", ScalarInteger(i32));
     if (TIFFGetField(tiff, TIFFTAG_IMAGEDEPTH, &i32))
 	setAttr(res, "depth", ScalarInteger(i32));
     if (TIFFGetField(tiff, TIFFTAG_BITSPERSAMPLE, &i16))
@@ -55,6 +59,16 @@ static void TIFF_add_info(TIFF *tiff, SEXP res) {
 	    setAttr(res, "planar.config", mkString(uv));
 	}
     }
+
+    if (TIFFGetField(tiff, TIFFTAG_ROWSPERSTRIP, &i32))
+        setAttr(res, "rows.per.strip", ScalarInteger(i32));
+
+    if (TIFFGetField(tiff, TIFFTAG_TILEWIDTH, &i32)) {
+        setAttr(res, "tile.width", ScalarInteger(i32));
+        TIFFGetField(tiff, TIFFTAG_TILELENGTH, &i32);
+        setAttr(res, "tile.length", ScalarInteger(i32));
+    }
+
     if (TIFFGetField(tiff, TIFFTAG_COMPRESSION, &i16)) {
 	char uv[24];
 	const char *name = 0;
@@ -82,6 +96,10 @@ static void TIFF_add_info(TIFF *tiff, SEXP res) {
 	setAttr(res, "x.resolution", ScalarReal(f));
     if (TIFFGetField(tiff, TIFFTAG_YRESOLUTION, &f))
 	setAttr(res, "y.resolution", ScalarReal(f));
+    if (TIFFGetField(tiff, TIFFTAG_XPOSITION, &f))
+	setAttr(res, "x.position", ScalarReal(f));
+    if (TIFFGetField(tiff, TIFFTAG_YPOSITION, &f))
+	setAttr(res, "y.position", ScalarReal(f));
     if (TIFFGetField(tiff, TIFFTAG_RESOLUTIONUNIT, &i16)) {
 	const char *name = "unknown";
 	switch (i16) {
@@ -142,16 +160,25 @@ static void TIFF_add_info(TIFF *tiff, SEXP res) {
 	setAttr(res, "color.space", mkString(name));
     }
 }
-    
-SEXP read_tiff(SEXP sFn, SEXP sNative, SEXP sAll, SEXP sConvert, SEXP sInfo, SEXP sIndexed, SEXP sOriginal) {
-    SEXP res = R_NilValue, multi_res = R_NilValue, multi_tail = R_NilValue, dim;
+
+SEXP read_tiff(SEXP sFn, SEXP sNative, SEXP sAll, SEXP sConvert, SEXP sInfo, SEXP sIndexed, SEXP sOriginal,
+	       SEXP sPayload) {
+    SEXP res = R_NilValue, multi_res = R_NilValue, multi_tail = R_NilValue, dim = R_NilValue;
     const char *fn;
-    int native = asInteger(sNative), all = (asInteger(sAll) == 1), n_img = 0,
+    int native = asInteger(sNative), all = (isLogical(sAll) && asInteger(sAll) > 0), n_img = 0,
 	convert = (asInteger(sConvert) == 1), add_info = (asInteger(sInfo) == 1),
-	indexed = (asInteger(sIndexed) == 1), original = (asInteger(sOriginal) == 1);
+	indexed = (asInteger(sIndexed) == 1), original = (asInteger(sOriginal) == 1),
+	info_only = (asInteger(sPayload) == 0);
     tiff_job_t rj;
     TIFF *tiff;
     FILE *f;
+    int *pick = (isInteger(sAll) ? INTEGER(sAll) : 0);
+    int picks = pick ? LENGTH(sAll) : 0;
+    SEXP pick_res = pick ? PROTECT(allocVector(VECSXP, picks)) : 0;
+
+    /* make sure people don't use vector logicals - they must use which() if that's what they want */
+    if (!((isLogical(sAll) && LENGTH(sAll) == 1) || isInteger(sAll)))
+	Rf_error("`all' must be either a scalar logical (TRUE/FALSE) or an integer vector");
 
     if (indexed && (convert || native))
 	Rf_error("indexed and native/convert cannot both be TRUE as they are mutually exclusive");
@@ -173,13 +200,69 @@ SEXP read_tiff(SEXP sFn, SEXP sNative, SEXP sAll, SEXP sConvert, SEXP sInfo, SEX
     if (!tiff)
 	Rf_error("Unable to open TIFF");
 
+    int cur_dir = 0; /* 1-based image number */
+    int nprot = 0;
     while (1) { /* loop over separate image in a directory if desired */
+	int pick_index = -1; /* not picked */
+	cur_dir++;
+
+	/* If sAll is a numeric vector, only read images referenced in it */
+	if (pick) {
+	    /* We don't use match() as it re-builds the hash table on every call. */
+	    int i = 0;
+	    while (i < picks) {
+		if (pick[i] == cur_dir) {
+		    pick_index = i;
+		    break;
+		}
+		i++;
+	    }
+
+	    if (pick_index == -1) { /* No match */
+		if (TIFFReadDirectory(tiff)) /* skip */
+		    continue;
+		else
+		    break;
+	    }
+	}
+
+
+	if (info_only) {
+	    /* dummy result object */
+	    res = PROTECT(allocVector(INTSXP, 0));
+	    TIFF_add_info(tiff, res);
+
+	    if (isLogical(sAll) && asInteger(sAll) == 0) {
+		UNPROTECT(1);
+		TIFFClose(tiff);
+		return res;
+	    }
+	    n_img++;
+	    if (pick_index >= 0)
+		SET_VECTOR_ELT(pick_res, pick_index, res);
+	    else {
+		if (multi_res == R_NilValue) {
+		    multi_tail = multi_res = CONS(res, R_NilValue);
+		    PROTECT(multi_res);
+		    nprot++;
+		} else {
+		    SEXP q = CONS(res, R_NilValue);
+		    SETCDR(multi_tail, q);
+		    multi_tail = q;
+		}
+	    }
+	    UNPROTECT(1);
+	    if (!TIFFReadDirectory(tiff))
+		break;
+	    continue;
+	}
+
 	uint32 imageWidth = 0, imageLength = 0, imageDepth;
 	uint32 tileWidth, tileLength;
 	uint32 x, y;
 	uint16 config, bps = 8, spp = 1, sformat = 1, out_spp;
 	tdata_t buf;
-	double *ra;
+	double *ra = 0;
 	uint16 *colormap[3] = {0, 0, 0};
 	int is_float = 0;
 
@@ -200,8 +283,10 @@ SEXP read_tiff(SEXP sFn, SEXP sNative, SEXP sAll, SEXP sConvert, SEXP sInfo, SEX
 	    if (colormap[2]) out_spp = 3;
 	    else if (colormap[1]) out_spp = 2;
 	}
-#if TIFF_DEBUG
-	Rprintf("image %d x %d x %d, tiles %d x %d, bps = %d, spp = %d (output %d), config = %d, colormap = %s\n", imageWidth, imageLength, imageDepth, tileWidth, tileLength, bps, spp, out_spp, config, colormap[0] ? "yes" : "no");
+#ifdef TIFF_DEBUG
+	Rprintf("image %d x %d x %d, tiles %d x %d, bps = %d, spp = %d (output %d), config = %d, colormap = %s,\n",
+		imageWidth, imageLength, imageDepth, tileWidth, tileLength, bps, spp, out_spp, config, colormap[0] ? "yes" : "no");
+	Rprintf("      float = %d\n", is_float);
 #endif
 	
 	if (native || convert) {
@@ -257,12 +342,13 @@ SEXP read_tiff(SEXP sFn, SEXP sNative, SEXP sAll, SEXP sConvert, SEXP sInfo, SEX
 		    TIFF_add_info(tiff, res);
 		UNPROTECT(1);
 	    } else {
+		SEXP R_ChannelsSymbol = Rf_install("channels");
 		dim = allocVector(INTSXP, 2);
 		INTEGER(dim)[0] = imageLength;
 		INTEGER(dim)[1] = imageWidth;
 		setAttrib(res, R_DimSymbol, dim);
 		setAttrib(res, R_ClassSymbol, mkString("nativeRaster"));
-		setAttrib(res, install("channels"), ScalarInteger(out_spp));
+		setAttrib(res, R_ChannelsSymbol, ScalarInteger(out_spp));
 		if (add_info)
 		    TIFF_add_info(tiff, res);
 		UNPROTECT(1);
@@ -272,26 +358,36 @@ SEXP read_tiff(SEXP sFn, SEXP sNative, SEXP sAll, SEXP sConvert, SEXP sInfo, SEX
 		return res;
 	    }
 	    n_img++;
-	    if (multi_res == R_NilValue) {
-		multi_tail = multi_res = CONS(res, R_NilValue);
-		PROTECT(multi_res);
-	    } else {
-		SEXP q = CONS(res, R_NilValue);
-		SETCDR(multi_tail, q);
-		multi_tail = q;
+	    if (pick_index >= 0)
+		SET_VECTOR_ELT(pick_res, pick_index, res);
+	    else {
+		if (multi_res == R_NilValue) {
+		    multi_tail = multi_res = CONS(res, R_NilValue);
+		    PROTECT(multi_res);
+		    nprot++;
+		} else {
+		    SEXP q = CONS(res, R_NilValue);
+		    SETCDR(multi_tail, q);
+		    multi_tail = q;
+		}
 	    }
 	    if (!TIFFReadDirectory(tiff))
 		break;
 	    continue;
-	}
-	
+	} /* end native || convert */
+
 	if (bps != 8 && bps != 16 && bps != 32 && ( bps != 12 || spp != 1 )) {
 	    TIFFClose(tiff);
 	    Rf_error("image has %d bits/sample which is unsupported in direct mode - use native=TRUE or convert=TRUE", bps);
 	}
 
-	if (sformat == SAMPLEFORMAT_INT)
-	    Rf_warning("tiff package currently only supports unsigned integer or float sample formats in direct mode, but the image contains signed integer format - it will be treated as unsigned (use native=TRUE or convert=TRUE to avoid this issue)");
+	if (original && is_float) {
+	    TIFFClose(tiff);
+	    Rf_error("as.is=TRUE is not supported for floating point images");
+	}
+
+	if (sformat == SAMPLEFORMAT_INT && !original)
+	    Rf_warning("tiff package currently only supports unsigned integer or float sample formats in direct mode, but the image contains signed integer format - it will be treated as unsigned (use as.is=TRUE, native=TRUE or convert=TRUE depending on your intent)");
 
 	/* FIXME: TIFF handle leak in case this fails */
 	res = allocVector((spp == 1 && (original || (indexed && colormap[0]))) ? INTSXP : REALSXP, imageWidth * imageLength * out_spp);
@@ -305,7 +401,7 @@ SEXP read_tiff(SEXP sFn, SEXP sNative, SEXP sAll, SEXP sConvert, SEXP sInfo, SEX
 	    tsize_t plane_offset = 0;
 	    x = 0; y = 0; 
 	    buf = _TIFFmalloc(TIFFStripSize(tiff));
-#if TIFF_DEBUG
+#ifdef TIFF_DEBUG
 	    Rprintf(" - %d x %d strips\n", TIFFNumberOfStrips(tiff), TIFFStripSize(tiff));
 #endif
 	    for (strip = 0; strip < TIFFNumberOfStrips(tiff); strip++) {
@@ -316,7 +412,7 @@ SEXP read_tiff(SEXP sFn, SEXP sNative, SEXP sAll, SEXP sConvert, SEXP sInfo, SEX
 			int *ia = original ? INTEGER(res) : 0;
 			if (bps == 12) step += 2;
 			for (i = 0; i < n; i += step) {
-			  unsigned int ci = 0, ci2;
+			    unsigned int ci = 0, ci2 = 0;
 			    const unsigned char *v = (const unsigned char*) buf + i;
 			    if (bps == 8) ci = v[0];
 			    else if (bps == 16) ci = ((const unsigned short int*)v)[0];
@@ -353,17 +449,17 @@ SEXP read_tiff(SEXP sFn, SEXP sNative, SEXP sAll, SEXP sConvert, SEXP sInfo, SEX
 			}
 		    } else if (colormap[0] || original) { /* indexed requested */
 			tsize_t i, step = bps / 8;
-			int *ia = INTEGER(res);
+			int *ia = INTEGER(res), ix_base = original ? 0 : 1;
 			if (bps == 12) step = 3;
 			for (i = 0; i < n; i += step) {
 			    int val = NA_INTEGER;
 			    const unsigned char *v = (const unsigned char*) buf + i;
-			    if (bps == 8) val = 1 + v[0];
-			    else if (bps == 16) val = 1 + ((const unsigned short int*)v)[0];
-			    else if (bps == 32) val = 1 + ((const unsigned int*)v)[0];
+			    if (bps == 8) val = ix_base + v[0];
+			    else if (bps == 16) val = ix_base + ((const unsigned short int*)v)[0];
+			    else if (bps == 32) val = ix_base + ((const unsigned int*)v)[0];
 			    else if (bps == 12) {
-				ia[imageLength * x++ + y] = DE12A(v);
-				val = DE12B(v);
+				ia[imageLength * x++ + y] = ix_base + DE12A(v);
+				val = ix_base + DE12B(v);
 			    }
 			    ia[imageLength * x++ + y] = val;
 			    if (x >= imageWidth) {
@@ -442,15 +538,74 @@ SEXP read_tiff(SEXP sFn, SEXP sNative, SEXP sAll, SEXP sConvert, SEXP sInfo, SEX
 		    }
 		}
 	    }
-	} else {
-	    Rf_error("tile-based images are currently supported only with native=TRUE or convert=TRUE");
-	    buf = _TIFFmalloc(TIFFTileSize(tiff));
+	} else { /* tiled image */
+	    if (indexed || colormap[0] || bps == 12)
+		Rf_error("Indexed and 12-bit tiled images are not supported.");
 	    
+	    if (spp > 1 && config != PLANARCONFIG_CONTIG)
+		Rf_error("Planar format tiled images are not supported");
+
+#ifdef TIFF_DEBUG
+	    Rprintf(" - %d x %d tiles\n", TIFFNumberOfTiles(tiff), TIFFTileSize(tiff));
+#endif
+	    x = 0; y = 0;
+	    buf = _TIFFmalloc(TIFFTileSize(tiff));
+
 	    for (y = 0; y < imageLength; y += tileLength)
-		for (x = 0; x < imageWidth; x += tileWidth)
-		    TIFFReadTile(tiff, buf, x, y, 0 /*depth*/, 0 /*plane*/);
+		for (x = 0; x < imageWidth; x += tileWidth) {
+		    tsize_t n = TIFFReadTile(tiff, buf, x, y, 0 /*depth*/, 0 /*plane*/);
+		    if (spp == 1) {
+			/* config doesn't matter for spp == 1 */
+			/* direct gray */
+			tsize_t i, step = bps / 8;
+			uint32 xoff=0, yoff=0;
+			for (i = 0; i < n; i += step) {
+			    double val = NA_REAL;
+			    const unsigned char *v = (const unsigned char*) buf + i;
+			    if (bps == 8) val = ((double) v[0]) / 255.0;
+			    else if (bps == 16) val = ((double) ((const unsigned short int*)v)[0]) / 65535.0;
+			    else if (bps == 32) {
+				if (is_float) val = (double) ((const float*)v)[0];
+				else val = ((double) ((const unsigned int*)v)[0]) / 4294967296.0;
+			    }
+			    if (x + xoff < imageWidth && y + yoff < imageLength)
+				ra[imageLength * (x + xoff) + y + yoff] = val;
+			    xoff++;
+			    if (xoff >= tileWidth) {
+				xoff -= tileWidth;
+				yoff++;
+			    }
+			}
+		    } else if (config == PLANARCONFIG_CONTIG) { /* interlaced */
+			tsize_t i, j, step = spp * bps / 8;
+			uint32 xoff=0, yoff=0;
+			for (i = 0; i < n; i += step) {
+			    const unsigned char *v = (const unsigned char*) buf + i;
+			    if (x + xoff < imageWidth && y + yoff < imageLength) {
+				if (bps == 8) {
+				    for (j = 0; j < spp; j++)
+					ra[(imageLength * imageWidth * j) + imageLength * (x + xoff) + y + yoff] = ((double) v[j]) / 255.0;
+				} else if (bps == 16) {
+				    for (j = 0; j < spp; j++)
+					ra[(imageLength * imageWidth * j) + imageLength * (x + xoff) + y + yoff] = ((double) ((const unsigned short int*)v)[j]) / 65535.0;
+				} else if (bps == 32 && !is_float) {
+				    for (j = 0; j < spp; j++)
+					ra[(imageLength * imageWidth * j) + imageLength * (x + xoff) + y + yoff] = ((double) ((const unsigned int*)v)[j]) / 4294967296.0;
+				} else if (bps == 32 && is_float) {
+				    for (j = 0; j < spp; j++)
+					ra[(imageLength * imageWidth * j) + imageLength * (x + xoff) + y + yoff] = (double) ((const float*)v)[j];
+				}
+			    }
+			    xoff++;
+			    if (xoff >= tileWidth) {
+				xoff -= tileWidth;
+				yoff++;
+			    }
+			} /* for i */
+		    } /* PLANARCONFIG_CONTIG */
+		} /* for x */
 	}
-	
+
 	_TIFFfree(buf);
 
 	PROTECT(res);
@@ -475,25 +630,37 @@ SEXP read_tiff(SEXP sFn, SEXP sNative, SEXP sAll, SEXP sConvert, SEXP sInfo, SEX
 	    TIFF_add_info(tiff, res);
 	UNPROTECT(1);
 	
-	if (!all) {
+	if (isLogical(sAll) && asInteger(sAll) == 0) {
 	    TIFFClose(tiff);
 	    return res;
 	}
 	n_img++;
-	if (multi_res == R_NilValue) {
-	    multi_tail = multi_res = CONS(res, R_NilValue);
-	    PROTECT(multi_res);
-	} else {
-	    SEXP q = CONS(res, R_NilValue);
-	    SETCDR(multi_tail, q);
-	    multi_tail = q;
+	if (pick_index >= 0)
+	    SET_VECTOR_ELT(pick_res, pick_index, res);
+	else {
+	    if (multi_res == R_NilValue) {
+		multi_tail = multi_res = CONS(res, R_NilValue);
+		PROTECT(multi_res);
+		nprot++;
+	    } else {
+		SEXP q = CONS(res, R_NilValue);
+		SETCDR(multi_tail, q);
+		multi_tail = q;
+	    }
 	}
 	if (!TIFFReadDirectory(tiff))
 	    break;
     }
     TIFFClose(tiff);
+    /* if picked, we already have the result list */
+    if (pick) {
+	UNPROTECT(nprot + 1 /* pick_res is the +1 */);
+	return pick_res;
+    }
+
     /* convert LISTSXP into VECSXP */
     PROTECT(res = allocVector(VECSXP, n_img));
+    nprot++;
     {
 	int i = 0;
 	while (multi_res != R_NilValue) {
@@ -502,6 +669,6 @@ SEXP read_tiff(SEXP sFn, SEXP sNative, SEXP sAll, SEXP sConvert, SEXP sInfo, SEX
 	    multi_res = CDR(multi_res);
 	}
     }
-    UNPROTECT(2);
+    UNPROTECT(nprot);
     return res;
 }
